@@ -1,5 +1,5 @@
 /*
-    Copyright (c) 2005-2018 Intel Corporation
+    Copyright (c) 2005-2020 Intel Corporation
 
     Licensed under the Apache License, Version 2.0 (the "License");
     you may not use this file except in compliance with the License.
@@ -12,15 +12,25 @@
     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     See the License for the specific language governing permissions and
     limitations under the License.
-
-
-
-
 */
 
+#include "internal/_deprecated_header_message_guard.h"
+
+#if !defined(__TBB_show_deprecation_message_task_H) && defined(__TBB_show_deprecated_header_message)
+#define  __TBB_show_deprecation_message_task_H
+#pragma message("TBB Warning: tbb/task.h is deprecated. For details, please see Deprecated Features appendix in the TBB reference manual.")
+#endif
+
+#if defined(__TBB_show_deprecated_header_message)
+#undef __TBB_show_deprecated_header_message
+#endif
+
 #ifndef __TBB_task_H
 #define __TBB_task_H
 
+#define __TBB_task_H_include_area
+#include "internal/_warning_suppress_enable_notice.h"
+
 #include "tbb_stddef.h"
 #include "tbb_machine.h"
 #include "tbb_profiling.h"
@@ -49,7 +59,9 @@ namespace internal { //< @cond INTERNAL
         task* self;
         task& parent;
     public:
-        explicit allocate_additional_child_of_proxy( task& parent_ ) : self(NULL), parent(parent_) {}
+        explicit allocate_additional_child_of_proxy( task& parent_ ) : self(NULL), parent(parent_) {
+            suppress_unused_warning( self );
+        }
         task& __TBB_EXPORTED_METHOD allocate( size_t size ) const;
         void __TBB_EXPORTED_METHOD free( task& ) const;
     };
@@ -118,6 +130,11 @@ namespace internal {
     /** Should always be non-negative.  A signed type is used so that underflow can be detected. */
     typedef intptr_t reference_count;
 
+#if __TBB_PREVIEW_RESUMABLE_TASKS
+    //! The flag to indicate that the wait task has been abandoned.
+    static const reference_count abandon_flag = reference_count(1) << (sizeof(reference_count)*CHAR_BIT - 2);
+#endif
+
     //! An id as used for specifying affinity.
     typedef unsigned short affinity_id;
 
@@ -162,6 +179,12 @@ namespace internal {
         void __TBB_EXPORTED_METHOD free( task& ) const;
     };
 
+#if __TBB_PREVIEW_CRITICAL_TASKS
+    // TODO: move to class methods when critical task API becomes public
+    void make_critical( task& t );
+    bool is_critical( task& t );
+#endif
+
     //! Memory prefix to a task object.
     /** This class is internal to the library.
         Do not reference it directly, except within the library itself.
@@ -187,6 +210,10 @@ namespace internal {
         friend class internal::allocate_child_proxy;
         friend class internal::allocate_continuation_proxy;
         friend class internal::allocate_additional_child_of_proxy;
+#if __TBB_PREVIEW_CRITICAL_TASKS
+        friend void make_critical( task& );
+        friend bool is_critical( task& );
+#endif
 
 #if __TBB_TASK_ISOLATION
         //! The tag used for task isolation.
@@ -211,7 +238,7 @@ namespace internal {
             thread-specific pools. */
         scheduler* origin;
 
-#if __TBB_TASK_PRIORITY
+#if __TBB_TASK_PRIORITY || __TBB_PREVIEW_RESUMABLE_TASKS
         union {
 #endif /* __TBB_TASK_PRIORITY */
         //! Obsolete. The scheduler that owns the task.
@@ -222,9 +249,16 @@ namespace internal {
 #if __TBB_TASK_PRIORITY
         //! Pointer to the next offloaded lower priority task.
         /** Used to maintain a list of offloaded tasks inside the scheduler. **/
-        task* next_offloaded;
+        tbb::task* next_offloaded;
+#endif
+
+#if __TBB_PREVIEW_RESUMABLE_TASKS
+        //! Pointer to the abandoned scheduler where the current task is waited for.
+        scheduler* abandoned_scheduler;
+#endif
+#if __TBB_TASK_PRIORITY || __TBB_PREVIEW_RESUMABLE_TASKS
         };
-#endif /* __TBB_TASK_PRIORITY */
+#endif /* __TBB_TASK_PRIORITY || __TBB_PREVIEW_RESUMABLE_TASKS */
 
         //! The task whose reference count includes me.
         /** In the "blocking style" of programming, this field points to the parent task.
@@ -274,6 +308,10 @@ namespace internal {
 #if __TBB_TASK_PRIORITY
 namespace internal {
     static const int priority_stride_v4 = INT_MAX / 4;
+#if __TBB_PREVIEW_CRITICAL_TASKS
+    // TODO: move into priority_t enum when critical tasks become public feature
+    static const int priority_critical = priority_stride_v4 * 3 + priority_stride_v4 / 3 * 2;
+#endif
 }
 
 enum priority_t {
@@ -294,6 +332,7 @@ enum priority_t {
 
 class task_scheduler_init;
 namespace interface7 { class task_arena; }
+using interface7::task_arena;
 
 //! Used to form groups of tasks
 /** @ingroup task_scheduling
@@ -320,7 +359,7 @@ class task_group_context : internal::no_copy {
 private:
     friend class internal::generic_scheduler;
     friend class task_scheduler_init;
-    friend class interface7::task_arena;
+    friend class task_arena;
 
 #if TBB_USE_CAPTURED_EXCEPTION
     typedef tbb_exception exception_container_type;
@@ -420,7 +459,7 @@ private:
     intptr_t my_priority;
 #endif /* __TBB_TASK_PRIORITY */
 
-    //! Decription of algorithm for scheduler based instrumentation.
+    //! Description of algorithm for scheduler based instrumentation.
     internal::string_index my_name;
 
     //! Trailing padding protecting accesses to frequently used members from false sharing
@@ -471,7 +510,7 @@ public:
         init();
     }
 
-    // Custom constructor for instrumentation of tbb algorithm
+    // Custom constructor for instrumentation of TBB algorithm
     task_group_context ( internal::string_index name )
         : my_kind(bound)
         , my_version_and_traits(3 | default_traits)
@@ -529,10 +568,10 @@ public:
 
 #if __TBB_TASK_PRIORITY
     //! Changes priority of the task group
-    void set_priority ( priority_t );
+    __TBB_DEPRECATED_IN_VERBOSE_MODE void set_priority ( priority_t );
 
     //! Retrieves current priority of the current task group
-    priority_t priority () const;
+    __TBB_DEPRECATED_IN_VERBOSE_MODE priority_t priority () const;
 #endif /* __TBB_TASK_PRIORITY */
 
     //! Returns the context's trait
@@ -573,7 +612,7 @@ private:
 
 //! Base class for user-defined tasks.
 /** @ingroup task_scheduling */
-class task: __TBB_TASK_BASE_ACCESS interface5::internal::task_base {
+class __TBB_DEPRECATED_IN_VERBOSE_MODE task: __TBB_TASK_BASE_ACCESS interface5::internal::task_base {
 
     //! Set reference count
     void __TBB_EXPORTED_METHOD internal_set_ref_count( int count );
@@ -609,6 +648,10 @@ public:
 #if __TBB_RECYCLE_TO_ENQUEUE
         //! task to be scheduled for starvation-resistant execution
         ,to_enqueue
+#endif
+#if __TBB_PREVIEW_RESUMABLE_TASKS
+        //! a special task used to resume a scheduler.
+        ,to_resume
 #endif
     };
 
@@ -797,11 +840,24 @@ public:
 #if __TBB_TASK_PRIORITY
     //! Enqueue task for starvation-resistant execution on the specified priority level.
     static void enqueue( task& t, priority_t p ) {
-        __TBB_ASSERT( p == priority_low || p == priority_normal || p == priority_high, "Invalid priority level value" );
+#if __TBB_PREVIEW_CRITICAL_TASKS
+        __TBB_ASSERT(p == priority_low || p == priority_normal || p == priority_high
+                     || p == internal::priority_critical, "Invalid priority level value");
+#else
+        __TBB_ASSERT(p == priority_low || p == priority_normal || p == priority_high, "Invalid priority level value");
+#endif
         t.prefix().owner->enqueue( t, (void*)p );
     }
 #endif /* __TBB_TASK_PRIORITY */
 
+    //! Enqueue task in task_arena
+    //! The implementation is in task_arena.h
+#if __TBB_TASK_PRIORITY
+    inline static void enqueue( task& t, task_arena& arena, priority_t p = priority_t(0) );
+#else
+    inline static void enqueue( task& t, task_arena& arena);
+#endif
+
     //! The innermost task being executed or destroyed by the current thread at the moment.
     static task& __TBB_EXPORTED_FUNC self();
 
@@ -830,6 +886,24 @@ public:
         return (prefix().extra_state & 0x80)!=0;
     }
 
+    //! True if the task was enqueued
+    bool is_enqueued_task() const {
+        // es_task_enqueued = 0x10
+        return (prefix().extra_state & 0x10)!=0;
+    }
+
+#if __TBB_PREVIEW_RESUMABLE_TASKS
+    //! Type that defines suspension point
+    typedef void* suspend_point;
+
+    //! Suspend current task execution
+    template <typename F>
+    static void suspend(F f);
+
+    //! Resume specific suspend point
+    static void resume(suspend_point tag);
+#endif
+
     //------------------------------------------------------------------------
     // Debugging
     //------------------------------------------------------------------------
@@ -840,10 +914,18 @@ public:
     //! The internal reference count.
     int ref_count() const {
 #if TBB_USE_ASSERT
+#if __TBB_PREVIEW_RESUMABLE_TASKS
+        internal::reference_count ref_count_ = prefix().ref_count & ~internal::abandon_flag;
+#else
         internal::reference_count ref_count_ = prefix().ref_count;
+#endif
         __TBB_ASSERT( ref_count_==int(ref_count_), "integer overflow error");
 #endif
+#if __TBB_PREVIEW_RESUMABLE_TASKS
+        return int(prefix().ref_count & ~internal::abandon_flag);
+#else
         return int(prefix().ref_count);
+#endif
     }
 
     //! Obsolete, and only retained for the sake of backward compatibility. Always returns true.
@@ -896,10 +978,10 @@ public:
 
 #if __TBB_TASK_PRIORITY
     //! Changes priority of the task group this task belongs to.
-    void set_group_priority ( priority_t p ) {  prefix().context->set_priority(p); }
+    __TBB_DEPRECATED void set_group_priority ( priority_t p ) {  prefix().context->set_priority(p); }
 
     //! Retrieves current priority of the task group this task belongs to.
-    priority_t group_priority () const { return prefix().context->priority(); }
+    __TBB_DEPRECATED priority_t group_priority () const { return prefix().context->priority(); }
 
 #endif /* __TBB_TASK_PRIORITY */
 
@@ -920,11 +1002,44 @@ private:
     internal::task_prefix& prefix( internal::version_tag* = NULL ) const {
         return reinterpret_cast<internal::task_prefix*>(const_cast<task*>(this))[-1];
     }
+#if __TBB_PREVIEW_CRITICAL_TASKS
+    friend void internal::make_critical( task& );
+    friend bool internal::is_critical( task& );
+#endif
 }; // class task
 
+#if __TBB_PREVIEW_CRITICAL_TASKS
+namespace internal {
+inline void make_critical( task& t ) { t.prefix().extra_state |= 0x8; }
+inline bool is_critical( task& t ) { return bool((t.prefix().extra_state & 0x8) != 0); }
+} // namespace internal
+#endif /* __TBB_PREVIEW_CRITICAL_TASKS */
+
+#if __TBB_PREVIEW_RESUMABLE_TASKS
+namespace internal {
+    template <typename F>
+    static void suspend_callback(void* user_callback, task::suspend_point tag) {
+        // Copy user function to a new stack to avoid a race when the previous scheduler is resumed.
+        F user_callback_copy = *static_cast<F*>(user_callback);
+        user_callback_copy(tag);
+    }
+    void __TBB_EXPORTED_FUNC internal_suspend(void* suspend_callback, void* user_callback);
+    void __TBB_EXPORTED_FUNC internal_resume(task::suspend_point);
+    task::suspend_point __TBB_EXPORTED_FUNC internal_current_suspend_point();
+}
+
+template <typename F>
+inline void task::suspend(F f) {
+    internal::internal_suspend((void*)internal::suspend_callback<F>, &f);
+}
+inline void task::resume(suspend_point tag) {
+    internal::internal_resume(tag);
+}
+#endif
+
 //! task that does nothing.  Useful for synchronization.
 /** @ingroup task_scheduling */
-class empty_task: public task {
+class __TBB_DEPRECATED_IN_VERBOSE_MODE empty_task: public task {
     task* execute() __TBB_override {
         return NULL;
     }
@@ -935,6 +1050,7 @@ namespace internal {
     template<typename F>
     class function_task : public task {
 #if __TBB_ALLOW_MUTABLE_FUNCTORS
+        // TODO: deprecated behavior, remove
         F my_func;
 #else
         const F my_func;
@@ -955,7 +1071,7 @@ namespace internal {
 //! A list of children.
 /** Used for method task::spawn_children
     @ingroup task_scheduling */
-class task_list: internal::no_copy {
+class __TBB_DEPRECATED_IN_VERBOSE_MODE task_list: internal::no_copy {
 private:
     task* first;
     task** next_ptr;
@@ -968,7 +1084,7 @@ public:
     //! Destroys the list, but does not destroy the task objects.
     ~task_list() {}
 
-    //! True if list if empty; false otherwise.
+    //! True if list is empty; false otherwise.
     bool empty() const {return !first;}
 
     //! Push task onto back of list.
@@ -1067,4 +1183,7 @@ inline void operator delete( void* task, const tbb::internal::allocate_additiona
     p.free( *static_cast<tbb::task*>(task) );
 }
 
+#include "internal/_warning_suppress_disable_notice.h"
+#undef __TBB_task_H_include_area
+
 #endif /* __TBB_task_H */
